#!/usr/bin/perl

#
# Reptor - Axent/Raptor firewall logfile analysis and reporting utility.
#
# Copyright (C) 1999,2000 Alex Howansky
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; see the file COPYING. If not, write to
# the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.
#

require 5.004;
use Getopt::Long;
use Net::SMTP;
use Net::FTP;

$reptor_version = "0.97";

# define options
GetOptions(
  "config=s" => \$config,
  "date=s" => \$date,
  "dir=s" => \$dir,
  "help" => \$help,
  "ignore" => \$ignore,
  "log=s" => \$log,
  "verify" => \$verify,
  "version" => \$version
);

version() if ($version);

# guess directory if not specified
$directory ||= $ENV{"EAGLEDIR"} ?
  $ENV{"EAGLEDIR"} . "/sg/oldlogs" :
  $^O eq "MSWin32" ?
    "c:/raptor/eagle/sg/oldlogs" :
    "/var/adm/sg/oldlogs";

# figure out what yesterday was
($year,$month,$day) = (localtime(time - 86400))[5,4,3];
$yesterday = sprintf "%4d%02d%02d", 1900 + $year, $month + 1, $day;

usage() if ($help);

# process configuration file
config($config || "reptor.cfg");

# command line overrides configuration file
$dir ||= $directory;

# figure out what logfile to process
unless ($log) {

  # default to yesterday unless overridden on the command line
  $file = $date ? "logfile.$date" : "logfile.$yesterday";

  # remotelog has precedence over local files
  if ($remotelog) {
    $log = $remotelog_path . "remotelogfile $remotelog $file |"
  } else {
    if ($uncompress) {
      $log = $uncompress . " " . $dir . "/" . $file . "." .$extension . " |";
    } else {
      $log = $dir . "/" . $file;
    }
  }

}

# open the logfile
unless (open LOG, $log) {
  # otherwise try the v4 naming style
  $log =~ s/logfile\.(?:19|20)([0-9]{6})/logfile\.$1/;
  open LOG, $log or die "ERROR: Unable to open logfile.\n";
}

if ($verify) {
  if ($remotelog) {
    print "Remotelog called successfully.\n";
  } else {
    if ($uncompress) {
      print "Uncompression program called successfully.\n";
    } else {
      print "Logfile opened successfully.\n";
    }
  }
}

%monthnum = (
  "Jan" => "01", "Feb" => "02", "Mar" => "03",
  "Apr" => "04", "May" => "05", "Jun" => "06",
  "Jul" => "07", "Aug" => "08", "Sep" => "09",
  "Oct" => "10", "Nov" => "11", "Dec" => "12"
);

@monthname = (
  "January", "February", "March", "April", "May", "June",
  "July", "August", "September", "October", "November", "December"
);

%linkalias = (
  "ftp" => "ftp",
  "ftp-data" => "ftp",
  "http" => "http",
  "http-https" => "https",
  "smtp" => "mailto",
  "109/tcp" => "mailto",
  "110/tcp" => "mailto",
  "119/tcp" => "news",
  "143/tcp" => "mailto"
);

@message_severity = (
  "Information",
  "Notice",
  "Warning",
  "Error",
  "Alert",
  "Critical",
  "Emergency",
);

%message_description = (
  "101" => "Raptor Network Security Management System starting up",
  "102" => "Shutdown command received",
  "103" => "Closing connection",
  "104" => "Re-read of new config file successful",
  "105" => "Connection for incoming to outgoing",
  "106" => "Path to destination",
  "107" => "Closing log file",
  "108" => "Starting new log file",
  "109" => "Re-reading configuration file",
  "110" => "User authenticated",
  "112" => "Rule expire, re-scanning rules",
  "115" => "Remote management connection",
  "116" => "Remote management completed",
  "117" => "Daemon starting",
  "118" => "Daemon exiting",
  "119" => "Request did not complete successfully",
  "120" => "Application information",
  "121" => "Statistics",
  "122" => "Daemon listening on port",
  "125" => "ISAKMP SA established with peer",
  "126" => "Protocol SA established with peer",
  "127" => "Connection request",
  "128" => "Finjan information",
  "129" => "Trace",
  "201" => "Access denied for IP address",
  "202" => "Access denied for user",
  "203" => "Password changed",
  "204" => "Password added",
  "209" => "VPN packet not valid for tunnel",
  "211" => "VPN authentication failed",
  "212" => "IP address not allowed on tunnel",
  "213" => "IP packet not allowed on implicit tunnel",
  "214" => "IP packet dropped",
  "215" => "VPN packet does not match any defined tunnel",
  "216" => "Access denied for IP address",
  "217" => "Cannot lookup hostname",
  "218" => "Invalid protocol",
  "219" => "Can't parse URL",
  "220" => "Local web server can't handle request",
  "221" => "Possible spoofed IP packet dropped",
  "222" => "VPN packet error",
  "223" => "VPN packet error",
  "224" => "User count is over limit",
  "225" => "Possible spoofed IP packet dropped",
  "226" => "IP packet dropped [restricted port]",
  "227" => "VPN packet dropped",
  "228" => "Can't connect to destination",
  "229" => "IP packet dropped [filter]",
  "230" => "Packet not allowed on tunnel",
  "231" => "Started on connection",
  "232" => "Sending original packet",
  "233" => "IP packet dropped [input packet filter]",
  "234" => "Network error",
  "235" => "NAT address allocation failed",
  "236" => "Cannot lookup active Protocol SA record",
  "237" => "Failed to connect to destination",
  "238" => "Application notice",
  "239" => "Sending TCP Reset, port not allowed",
  "301" => "Internal warning",
  "303" => "Service is already running",
  "306" => "Overlapping time range",
  "308" => "Can't lookup hostname",
  "309" => "Warning in config file",
  "310" => "Can't verify reverse address",
  "311" => "Can't verify ethernet address",
  "312" => "Bogus response to hostname lookup",
  "313" => "Access denied to user [bad password]",
  "314" => "Warning in user file [blank user name]",
  "315" => "Warning in user file [non-numeric user id]",
  "316" => "Warning in user file [non-numeric group id]",
  "317" => "Warning in user file [blank group name]",
  "318" => "Warning in user file [non-numeric group id]",
  "323" => "Readeagle already running",
  "333" => "No rules for firewall in config file",
  "334" => "Denied access to protocol command",
  "335" => "VPN packet dropped [VPN is not enabled]",
  "336" => "VPN packet dropped [invalid format or length]",
  "338" => "VPN packet dropped [UDP disabled]",
  "341" => "Child process stopped unexpectedly",
  "342" => "Child process exited with an unexpected error",
  "343" => "Software problem",
  "344" => "Non-transparent call",
  "345" => "Bad DNS information",
  "347" => "Possible port scan detected",
  "401" => "Internal error",
  "402" => "Can't get config file",
  "403" => "Can't read config file",
  "404" => "Error in config file",
  "405" => "Re-read of config file failed",
  "406" => "Cannot open audio file",
  "407" => "Can't open traceroute lockfile",
  "408" => "Not a valid audio file",
  "409" => "Audio file sample rate not available",
  "410" => "Filename encoding not available",
  "411" => "Unable to open notify schedule",
  "412" => "Unrecognized transport",
  "413" => "Have config file errors but continuing",
  "415" => "Syntax error in config file date/time string",
  "416" => "Syntax error in config file expression string",
  "418" => "Can't lookup host",
  "420" => "Can't open config file",
  "423" => "Bad protocol",
  "424" => "Can't use TCP port",
  "425" => "Can't lookup service",
  "430" => "EagleLAN can't lookup firewall hostname",
  "431" => "EagleLAN can't read firewall location",
  "432" => "Bad hostname",
  "433" => "Can't connect to host",
  "434" => "Error reading config file",
  "435" => "Can't execute service to read config file",
  "440" => "EagleLAN contact failed",
  "441" => "Parameter error",
  "442" => "VPN: Could not attach to data link driver",
  "444" => "Error in password file",
  "445" => "Can't read password file",
  "446" => "No entry for host in password file",
  "450" => "Management of service failed",
  "451" => "Bad port in config file",
  "452" => "Can't lookup proxy name",
  "454" => "Can't open file",
  "455" => "Child process exited under abnormal conditions",
  "456" => "HTTPS service not supported",
  "457" => "Application error",
  "501" => "Suspicious activity monitor threshold triggered",
  "502" => "Ethernet address mismatch for host",
  "503" => "Reverse address doesn't match for host",
  "504" => "Unknown entity connected to readeagle",
  "505" => "Unauthorized process killed",
  "506" => "Unauthorized user logged off",
  "510" => "Incorrect data checksum",
  "511" => "Incorrect challenge",
  "512" => "Unauthorized remote connect attempt",
  "513" => "Saved SMTP trace file",
  "514" => "Unauthorized protocol",
  "515" => "User attempted to connect to control port",
  "516" => "DNSd error",
  "523" => "Hostile Java applet blocked by SurfinGate",
  "601" => "Child process died unexpectedly",
  "602" => "Child process returned an unexpected error",
  "603" => "Fork failed",
  "605" => "Can't execute service",
  "606" => "Failed to notify",
  "607" => "Daemon exited on signal",
  "609" => "Syslog daemon not running",
  "610" => "Internal error",
  "611" => "User count limit reached",
  "612" => "Service error",
  "701" => "Can't allocate memory",
  "702" => "Quitting because of config errors",
  "704" => "Expriation date reached",
  "705" => "Invalid license key",
  "706" => "Module not licensed",
  "707" => "Service not installed",
  "708" => "Service is corrupt"
);

# attempt to extract date information from first line
if (!$ignore) {
  $_ = <LOG>;
  if (/^(...) (..).+, Year is ([0-9]{4}), /) {
    $datetitle = "$monthname[$monthnum{$1} - 1] $2, $3";
    $filename = $3 . $monthnum{$1} . $2 . ".html";
  } else {
    die "Error: Logfile does not exist or has been cut. Specify --ignore if cut.\n";
  }
} else {
  $filename = "reptor.html";
}

if ($verify) {
  close LOG;
  print "Everything seems to be ok.\n";
  exit;
}

# create a local copy of the logfile
if ($save_logfile) {
  $save_logfile =~ s/([^\/\\])$/$1\//;
  $save_name = $save_logfile . ($file ? $file : ($log =~ /([^\/\\]+)$/)[0]);
  if (-r $save_name) {
    die "ERROR: Filename for copy of logfile is already in use.\n";
  }
  open SAVE, ">$save_name" or
    die "ERROR: Unable to create copy of logfile in $save_name.\n";
  print SAVE;
}

# open the output file
if ($output_how =~ /^(ftp|file|mail)$/) {
  if ($output_how eq "file") {
    if ($output_to =~ /[\/\\]$/) {
      open OUTPUT, ">$output_to/$filename" or
        die "ERROR: Unable to open output file $output_to.\n";
    } else {
      open OUTPUT, ">$output_to" or
        die "ERROR: Unable to open output file $output_to.\n";
    }
  } else {
    open OUTPUT, ">reptor.$$" or
      die "ERROR: Unable to create temporary file.\n";
  }
  select OUTPUT;
}

# print HTML header
print
  "<html>\n",
  "<head>\n",
  "<title>Firewall activity",
  $datetitle ? " for $datetitle" : "",
  "</title>\n";

# insert style sheet if available
if ($style_sheet) {
  if (! -r $style_sheet) {
    die "Error: Unable to open style sheet $style_sheet.\n";
  }
  print
    "<style type=\"text/css\">\n",
    "<!--\n";
  open CSS, $style_sheet or die "ERROR: Unable to open style sheet.\n";
  while (<CSS>) { print; }
  close CSS;
  print
    "-->\n",
    "</style>\n";
}

print
  "<meta name=\"description\" content=\"Firewall activity",
  $datetitle ? " for $datetitle" : "", "\">\n",
  "<meta name=\"keywords\" content=\"reptor raptor firewall",
  $datetitle ? " $datetitle" : "", "\">\n",
  "</head>\n",
  "<body>\n",
  "<h1>$title", $datetitle ? "<br>$datetitle" : "", "</h1>\n";

# print interface filter list
if (%filter) {
  print
    "<p>\n",
    "<dl>\n",
    "<dt>Interfaces filtered:</dt>\n";
  for ($index = 0; $index <= $#{$filter{"source"}}; $index++) {
    print
      "<dd>",
      "From ",
      $filter{"source"}[$index],
      " to ",
      $filter{"destination"}[$index],
      "</dd>\n";
  }
  print
    "</dl>\n",
    "</p>\n";
}

# print headers for alerts section
if ($alerts) {
  if ($links) { makelinks("alerts"); }
  print
    "<p>\n",
    "<table $table>\n",
    "<tr><th align=\"left\" colspan=\"$columns\">Alerts</th></tr>\n",
    "<tr>\n";
  if ($show_reason) { print "<th>Reason</th>\n"; }
  print
    "<th>Time</th>\n",
    "<th>Duration</th>\n",
    "<th>Protocol</th>\n",
    "<th>Bytes<br>Sent</th>\n",
    "<th>Bytes<br>Received</th>\n",
    "<th>Bytes<br>Total</th>\n";
  if ($show_interface) { print "<th>Source<br>Interface</th>\n"; }
  print "<th>Source<br>Host</th>\n";
  if ($show_interface) { print "<th>Destination<br>Interface</th>\n"; }
  print "<th>Destination<br>Host</th>\n";
  if ($show_user) { print "<th>User</th>\n"; }
  if ($show_auth) { print "<th>Authentication<br>Method</th>\n"; }
  if ($show_rule) { print "<th>Rule</th>\n"; }
  if ($show_id) { print "<th>ID</th>\n"; }
  if ($show_op) { print "<th>Op</th>\n"; }
  if ($show_result) { print "<th>Result</th>\n"; }
  if ($show_arg) { print "<th>Argument</th>\n"; }
  print "</tr>\n";
}

while (<LOG>) {

  # localize variables
  my(
    $user,$auth,$sent,$rcvd,$srcif,$src,$dstif,$dst,$total,
    $result,$id,$type,$time,$arg,$proto,$ok,$linkproto,$line,
    $newtime,$alerted,$search,$rule,$op,$domain,$duration,$hour
  );

  # make local copy of logfile
  if ($save_logfile) { print SAVE; }

  # parse out time and message type
  ($time,$type,$line) =
    /^
    [a-zA-Z]{3}[ ]     # month
    [0-9]{2}[ ]        # day
    ([0-9:]{8})        # hour:minute:second
    (?:\.[0-9]{3})?[ ] # microsecond (v6 only)
    [^ ]+[ ]           # firewall name
    [^ ]+[ ]           # subsystem & pid?
    ([0-9]+)[ ]        # message number
    (.+)               # rest of line
    $/x;

  # accumulate message summaries
  if (%message_summary && $type >= $message_summary{"limit"}) {
    $total_message{$type}++;
  }

  $_ = $line;

  # check for Raptor Mobile connections
  if (%mobile_summary && $type == 120 && /^nsetupd Info: Enabling tunnels for Mobile ([^ ]+) /) {
    $total_mobile{$1}++;
    $total_mobile++;
  }

  # only calculate stats for logfile entries that have type = 121
  next if ($type != 121);

  # only calculate stats for logfile entries that have duration > 0
  if (/^Statistics: duration=([^ ]+) /) {

    # duration of connection
    $duration = int($1 + 0.5);

    # source interface
    $srcif = (/ srcif=([^ ]+) /) ? $interface{$1} : "Unidentified";

    # destination interface
    $dstif = (/ dstif=([^ ]+) /) ? $interface{$1} : "Unidentified";

    # check for filtered interfaces before continuing
    if (%filter) {
      for ($index = 0; $index <= $#{$filter{"source"}}; $index++) {
        if (($filter{"source"}[$index] eq "any" ||
            $srcif eq $filter{"source"}[$index]) &&
            ($filter{"destination"}[$index] eq "any" ||
            $dstif eq $filter{"destination"}[$index])) {
          $ok = 1;
          last;
        }
      }
      next if (!$ok);
    }

    # bytes sent
    ($sent) = / sent=([0-9]+) /;

    # bytes received
    ($rcvd) = / rcvd=([0-9]+) /;

    # total bytes
    $total = $sent + $rcvd;

    # source host
    unless($src = lc((/ srcname=([^ ]+) /)[0])) {
      $src = (split(/\//,(/ src=([^ ]+) /)[0]))[0];
      if ($dns eq "all") {
        $src = lc(gethostbyaddr(pack("c4",split(/\./,$src)),2)) || "[$src]";
      }
    }

    # destination host
    unless($dst = lc((/ dstname=([^ ]+) /)[0])) {
      $dst = (split(/\//,(/ dst=([^ ]+) /)[0]))[0];
      if ($dns eq "all") {
        $dst = lc(gethostbyaddr(pack("c4",split(/\./,$dst)),2)) || "[$dst]";
      }
    }

    # protocol
    if (/ proto=([^ ]+) /) {
      $proto = $alias{$1} || $1;
      if ($links && $linkalias{$1}) { $linkproto = $linkalias{$1}; }
    }

    # user
    if (%user_summary || ($alerts && ($show_user || %user))) {
      if (/ user=([^ ]+) /) {
        $user = $1;
        # remove less than and greater than signs
        $user =~ s/^<//g;
        $user =~ s/>$//g;
      }
    }

    # process the entry for alert triggering
    if ($alerts) {

      # authentication method
      if ($show_auth) { ($auth) = / auth=([^ ]+) /; }

      # rule
      if ($show_rule) { ($rule) = / rule=([^ ]+) /; }

      # id
      if ($show_id) { ($id) = / id=([^ ]+) /; }

      # op
      if ($show_op) { ($op) = / op=([^\" ]+|\"[^\"]+\") /; }

      # result
      if ($show_result) { ($result) = / result=([^\" ]+|\"[^\"]+\") /; }

      # argument
      if ($show_arg) { ($arg) = / arg=([^ ]+) /; }

      # build string for word search
      if ($word) {
        if ($search_source) { $search .= " $src"; }
        if ($search_destination) { $search .= " $dst"; }
        if ($search_arg) { $search .= " $arg"; }
      }

      # determine if this entry triggers an alert condition
      # manually control short-circuiting to optimize speed
      # perform most computationally expensive checks last

      # check for all option
      if ($all) { $alerted = "All"; }

      # rating
      if ($ratings && / \(Denied due to ratings check\)$/) {
        $alerted = "Ratings";
      }

      # check for user
      if (!$alerted && $user{$proto}{$user}) { $alerted = "User"; }

      # check for duration
      if (!$alerted && $duration{$proto} && $duration >= $duration{$proto}) {
        $alerted = "Duration";
      }

      # check for volume
      if (!$alerted && $volume{$proto} && $total >= $volume{$proto}) {
        $alerted = "Volume";
      }

      # check for time of day
      if (!$alerted && $time{$proto}) {
        $newtime = substr($time,0,2) . substr($time,3,2);
        if (substr($time,6,2) > 30) { $newtime++; }
        for ($index = 0; $index <= $#{$time{$proto}{"begin"}}; $index++) {
          if ($newtime >= $time{$proto}{"begin"}[$index] &&
              $newtime <= $time{$proto}{"end"}[$index]) {
            $alerted = "Time";
            last;
          }
        }
      }

      # check for word
      if (!$alerted && $word && $search =~ $word) { $alerted = "Word"; }

      # print the entry if an alert triggered
      if ($alerted) {

        # accumulate number of alerts for this host
        if (%alert_summary) { $total_host_alerts{$src}++; }

        # make less than and greater than signs HTML friendly
        if ($show_arg && $arg) {
          $arg =~ s/</&lt;/g;
          $arg =~ s/>/&gt;/g;
        }

        print "<tr>\n";

        if ($highlight) { $class = $class eq "odd" ? "even" : "odd"; }

        if ($show_reason) {
          print "<td";
          if ($highlight) { print " class=\"$class\""; }
          print " align=\"left\">$alerted</td>\n";
        }

        print "<td";
        if ($highlight) { print " class=\"$class\""; }
        print " align=\"right\">$time</td>\n";
        print "<td";
        if ($highlight) { print " class=\"$class\""; }
        print " align=\"right\">", fduration($duration), "</td>\n";
        print "<td";
        if ($highlight) { print " class=\"$class\""; }
        print " align=\"left\"><nobr>$proto</nobr></td>\n";
        print "<td";
        if ($highlight) { print " class=\"$class\""; }
        print " align=\"right\"><nobr>", bytes($sent), "</nobr></td>\n";
        print "<td";
        if ($highlight) { print " class=\"$class\""; }
        print " align=\"right\"><nobr>", bytes($rcvd), "</nobr></td>\n";
        print "<td";
        if ($highlight) { print " class=\"$class\""; }
        print " align=\"right\"><nobr>", bytes($total), "</nobr></td>\n";

        if ($show_interface) {
          print "<td";
          if ($highlight) { print " class=\"$class\""; }
          print " align=\"left\"><nobr>$srcif</nobr></td>\n";
        }

        print "<td";
        if ($highlight) { print " class=\"$class\""; }
        print " align=\"left\"><nobr>", name($src), "</nobr></td>\n";

        if ($show_interface) {
          print "<td";
          if ($highlight) { print " class=\"$class\""; }
          print " align=\"left\"><nobr>$dstif</nobr></td>\n";
        }

        print "<td";
        if ($highlight) { print " class=\"$class\""; }
        print " align=\"left\"><nobr>";
        if ($links && $linkproto && $linkproto ne "mailto") {
          print
            "<a href=\"$linkproto://$dst\" target=\"_blank\">",
            name($dst),
            "</a>";
        } else {
          print name($dst);
        }
        print "</nobr></td>\n";

        if ($show_user) {
          print "<td";
          if ($highlight) { print " class=\"$class\""; }
          print " align=\"left\">";
          if ($user) {
            print "<nobr>";
            if ($links && $linkproto eq "mailto") {
              print "<a href=\"mailto:$user\">$user</a>";
            } else {
              print "$user";
            }
            print "</nobr>";
          }
          print "</td>\n";
        }

        if ($show_auth) {
          print "<td";
          if ($highlight) { print " class=\"$class\""; }
          print " align=\"left\"><nobr>$auth</nobr></td>\n";
        }

        if ($show_rule) {
          print "<td";
          if ($highlight) { print " class=\"$class\""; }
          print " align=\"right\">$rule</td>\n";
        }

        if ($show_id) {
          print "<td";
          if ($highlight) { print " class=\"$class\""; }
          print " align=\"right\">$id</td>\n";
        }

        if ($show_op) {
          print "<td";
          if ($highlight) { print " class=\"$class\""; }
          print " align=\"right\"><nobr>$op</nobr></td>\n";
        }

        if ($show_result) {
          print "<td";
          if ($highlight) { print " class=\"$class\""; }
          print " align=\"left\"><nobr>$result</nobr></td>\n";
        }

        if ($show_arg) {
          print "<td";
          if ($highlight) { print " class=\"$class\""; }
          print " align=\"left\">";
          if ($arg) {
            print "<nobr>";
            if ($links && $linkproto && $linkproto ne "mailto") {
              if ($arg =~ /^http:\/\//) {
                print "<a href=\"$arg\" target=\"_blank\">$arg</a>";
              } else {
                print "<a href=\"$linkproto://$dst$arg\" target=\"_blank\">$arg</a>";
              }
            } else {
              print "$arg";
            }
            print "</nobr>";
          }
          print "</td>\n";
        }

        print "</tr>\n";

      } # if ($alerted)

    } # if ($alerts)

    # accumulate total stats
    $total_hits++;
    $total_total += $total;
    $total_duration += $duration;

    # accumulate top level domain stats
    if (%domain_summary) {
      if ($src =~ /\.[a-z]/) {
        $domain = substr($src,rindex($src,".") + 1);
        $total_domain_hits{$domain}++;
        $total_domain_sent{$domain} += $sent;
        $total_domain_received{$domain} += $rcvd;
        $total_domain_total{$domain} += $total;
        $total_domain_duration{$domain} += $duration;
      }
      if ($dst =~ /\.[a-z]/) {
        $domain = substr($dst,rindex($dst,".") + 1);
        $total_domain_hits{$domain}++;
        $total_domain_sent{$domain} += $rcvd;
        $total_domain_received{$domain} += $sent;
        $total_domain_total{$domain} += $total;
        $total_domain_duration{$domain} += $duration;
      }
    }

    # accumulate time of day stats
    if (%hour_summary) {
      ($hour) = split(/:/,$time);
      $total_hour_hits{$hour}++;
      $total_hour_sent{$hour} += $sent;
      $total_hour_received{$hour} += $rcvd;
      $total_hour_total{$hour} += $total;
      $total_hour_duration{$hour} += $duration;
    }

    # accumulate interface stats
    if (%interface_summary) {
      $total_interface_hits{$srcif}++;
      $total_interface_hits{$dstif}++;
      $total_interface_sent{$srcif} += $sent;
      $total_interface_sent{$dstif} += $rcvd;
      $total_interface_received{$srcif} += $rcvd;
      $total_interface_received{$dstif} += $sent;
      $total_interface_duration{$srcif} += $duration;
      $total_interface_duration{$dstif} += $duration;
      $total_interface_total{$srcif} += $total;
      $total_interface_total{$dstif} += $total;
    }

    # accumulate host stats
    if (%host_summary) {
      $total_host_hits{$srcif}{$src}++;
      $total_host_hits{$dstif}{$dst}++;
      $total_host_sent{$srcif}{$src} += $sent;
      $total_host_sent{$dstif}{$dst} += $rcvd;
      $total_host_received{$srcif}{$src} += $rcvd;
      $total_host_received{$dstif}{$dst} += $sent;
      $total_host_total{$srcif}{$src} += $total;
      $total_host_total{$dstif}{$dst} += $total;
      $total_host_duration{$srcif}{$src} += $duration;
      $total_host_duration{$dstif}{$dst} += $duration;
    }

    # accumulate user stats
    if (%user_summary && $user) {
      $total_user_hits{$user}++;
      $total_user_sent{$user} += $sent;
      $total_user_received{$user} += $rcvd;
      $total_user_total{$user} += $total;
      $total_user_duration{$user} += $duration;
    }

    # accumulate protocol stats
    if (%protocol_summary) {
      $total_protocol_hits{$proto}++;
      $total_protocol_sent{$proto} += $sent;
      $total_protocol_received{$proto} += $rcvd;
      $total_protocol_total{$proto} += $total;
      $total_protocol_duration{$proto} += $duration;
    }

  } # if entry selected

} # while (<LOG>)

if ($alerts) { print "</table>\n</p>\n"; }

close LOG;

if ($save_logfile) {
  close SAVE;
  if ($compress) { system(split(/ /,$compress),$save_name); }
}

# only update history if we're processing yesterday's logfile
if ($history_file && $file eq "logfile.$yesterday") {
  open HISTORY, ">>$history_file"
    or die "ERROR: Unable to open history file.\n";
  print HISTORY "$yesterday,$total_hits,$total_total,$total_duration\n";
  close HISTORY;
}

if (%alert_summary) { &alert_summary; }

if ($totals) { &total_summary; }

if (%protocol_summary) { &protocol_summary; }

if (%hour_summary) { &hour_summary; }

if (%interface_summary) { &interface_summary; }

if (%host_summary) {
  if ($links) { makelinks("host_summary"); }
  foreach $index (sort keys %host_summary) { host_summary($index); }
}

if (%user_summary) { &user_summary; }

if (%domain_summary) { &domain_summary; }

if (%mobile_summary) { &mobile_summary; }

if (%message_summary) { &message_summary; }

if (%history_summary) {
  if ($links) { makelinks("history_summary"); }
  foreach $index (sort keys %history_summary) { history_summary($index); }
}

# finish HTML output
print
  "<p>\n",
  "Generated by ",
  "<a href=\"http://www.wankwood.com/reptor/\" target=\"_blank\">Reptor</a> version $reptor_version\n",
  "</p>\n",
  "</body>\n",
  "</html>\n";

# send off the output file
if ($output_how =~ /^(ftp|file|mail)$/) {
  close OUTPUT;

  if ($output_how eq "ftp") {
    my ($ftp_user,$ftp_pwd,$ftp_host,$ftp_file,$ftp);
    ($ftp_user,$ftp_pwd,$ftp_host,$ftp_file) = split(/,/,$output_to);
    if ($ftp_file =~ /[\/\\]$/) { $ftp_file .= $filename; }
    $ftp = Net::FTP->new($ftp_host);
    $ftp->pasv();
    $ftp->login($ftp_user,$ftp_pwd);
    $ftp->put("reptor.$$",$ftp_file);
    $ftp->quit;
    unlink "reptor.$$";
  }

  if ($output_how eq "mail") {
    my ($mail_date,$smtp);
    $mail_date = localtime(time);
    $smtp = Net::SMTP->new($mail_server);
    $smtp->mail($mail_from);
    $smtp->to($output_to);
    $smtp->data();
    $smtp->datasend("From: $mail_from\n");
    $smtp->datasend("To: $output_to\n");
    $smtp->datasend("Date: $mail_date\n");
    $smtp->datasend("Subject: Reptor report" .
      ($datetitle ? " for $datetitle" : "") . "\n");
    $smtp->datasend("MIME-Version: 1.0\n");
    $smtp->datasend("Content-type: text/html\n");
    $smtp->datasend("\n");
    open TEMP, "reptor.$$" or die "ERROR: Unable to open temporary file.\n";
    while (<TEMP>) { $smtp->datasend($_); }
    close TEMP;
    $smtp->dataend();
    $smtp->quit;
    unlink "reptor.$$";
  }
}

sub total_summary {
  my ($index,$class);
  if ($links) { makelinks("totals"); }
  print
    "<p>\n",
    "<table $table>\n",
    "<tr><th align=\"left\" colspan=\"2\">Totals</th></tr>\n",
    "<tr>\n",
    "<th align=\"right\">Hits</th>\n",
    "<td";
  if ($highlight) { print " class=\"odd\""; }
  print
    ">$total_hits</td>\n",
    "</tr>\n",
    "<tr>\n",
    "<th align=\"right\">Duration</th>\n",
    "<td";
  if ($highlight) { print " class=\"even\""; }
  print
    ">", fduration($total_duration), "</td>\n",
    "</tr>\n",
    "<tr>\n",
    "<th align=\"right\">Bytes</th>\n",
    "<td";
  if ($highlight) { print " class=\"odd\""; }
  print
    ">", bytes($total_total), "</td>\n",
    "</tr>\n";
  if (%host_summary) {
    foreach $index (sort keys %host_summary) {
      if ($highlight) { $class = $class eq "even" ? "odd" : "even"; }
      print
        "<tr>\n",
        "<th align=\"right\">Hosts [$index]</th>\n",
        "<td";
      if ($highlight) { print " class=\"$class\""; }
      print
        ">", scalar keys %{$total_host_hits{$index}}, "</td>\n",
        "</tr>\n",
    }
  }
  print
    "</table>\n",
    "</p>\n";
}

sub alert_summary {
  my ($num,$direction,$sort,$class,$index);
  $num = 0;
  if ($links) { makelinks("alert_summary"); }
  print
    "<p>\n",
    "<table $table>\n",
    "<tr><th align=\"left\" colspan=\"2\">Alert Summary</th></tr>\n",
    "<tr>\n",
    "<th>Host</th>\n",
    "<th>Alerts</th>\n",
    "</tr>\n";
  $direction = substr($alert_summary{"sort"},0,1);
  substr($alert_summary{"sort"},0,1) = "";
  if ($alert_summary{"sort"} eq "host") {
    $sort = 'sort keys %total_host_alerts';
  } else {
    $sort = 'sort { $total_host_alerts{$a} <=> $total_host_alerts{$b} } keys %total_host_alerts';
  }
  if ($direction eq "-") { $sort = 'reverse ' . $sort; }
  foreach $index (eval($sort)) {
    if ($highlight) { $class = $class eq "odd" ? "even" : "odd"; }
    print
      "<tr>\n",
      "<td";
    if ($highlight) { print " class=\"$class\""; }
    print
      " align=\"right\"><nobr>", name($index), "</nobr></td>\n",
      "<td";
    if ($highlight) { print " class=\"$class\""; }
    print
      " align=\"right\">$total_host_alerts{$index}</td>\n",
      "</tr>\n";
    last if ($alert_summary{"limit"} && ++$num >= $alert_summary{"limit"});
  }
  print
    "</table>\n",
    "</p>\n";
}

sub protocol_summary {
  my ($num,$direction,$sort,$class,$index);
  $num = 0;
  if ($links) { makelinks("protocol_summary"); }
  print
    "<p>\n",
    "<table $table>\n";
  if ($graphs) {
    print "<tr><th align=\"left\" colspan=\"8\">Protocol Summary</th></tr>\n";
  } else {
    print "<tr><th align=\"left\" colspan=\"7\">Protocol Summary</th></tr>\n";
  }
  print
    "<tr>\n",
    "<th>Protocol</th>\n",
    "<th>Hits</th>\n",
    "<th>Duration</th>\n",
    "<th>Bytes<br>Sent</th>\n",
    "<th>Bytes<br>Received</th>\n",
    "<th>Bytes<br>Total</th>\n";
  if ($graphs) {
    print "<th colspan=\"2\">Percent of Total Bytes</th>\n",
  } else {
    print "<th>Percent of<br>Total Bytes</th>\n",
  }
  print "</tr>\n";
  $direction = substr($protocol_summary{"sort"},0,1);
  substr($protocol_summary{"sort"},0,1) = "";
  if ($protocol_summary{"sort"} eq "protocol") {
    $sort = 'sort keys %total_protocol_hits';
  } else {
    $sort =
     'sort { $total_protocol_' . $protocol_summary{"sort"} .
     '{$a} <=> $total_protocol_' . $protocol_summary{"sort"} .
     '{$b} } keys %total_protocol_hits';
  }
  if ($direction eq "-") { $sort = 'reverse ' . $sort; }
  foreach $index (eval($sort)) {
    if ($highlight) { $class = $class eq "odd" ? "even" : "odd"; }
    print
      "<tr>\n",
      "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\"><nobr>$index</nobr></td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">$total_protocol_hits{$index}</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"left\">", fduration($total_protocol_duration{$index}), "</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_protocol_sent{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_protocol_received{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_protocol_total{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    printf " align=\"right\">%.2f</td>\n",
      100 * $total_protocol_total{$index} / $total_total;
    if ($graphs) {
      print "<td";
      if ($highlight) { print " class=\"$class\""; }
      print ">";
      printf
        "<img alt=\"\" src=\"pixel.gif\" width=\"%d\" height=\"12\"></td>",
        width($total_protocol_total{$index},$total_total);
    }
    print "</tr>\n";
    last if ($protocol_summary{"limit"} && ++$num >= $protocol_summary{"limit"});
  }
  print
    "</table>\n",
    "</p>\n";
}

sub interface_summary {
  my ($num,$direction,$sort,$class,$index);
  $num = 0;
  if ($links) { makelinks("interface_summary"); }
  print
    "<p>\n",
    "<table $table>\n";
  if ($graphs) {
    print "<tr><th align=\"left\" colspan=\"8\">Interface Summary</th></tr>\n";
  } else {
    print "<tr><th align=\"left\" colspan=\"7\">Interface Summary</th></tr>\n";
  }
  print
    "<tr>\n",
    "<th>Interface</th>\n",
    "<th>Hits</th>\n",
    "<th>Duration</th>\n",
    "<th>Bytes<br>Sent</th>\n",
    "<th>Bytes<br>Received</th>\n",
    "<th>Bytes<br>Total</th>\n";
  if ($graphs) {
    print "<th colspan=\"2\">Percent of Total Bytes</th>\n",
  } else {
    print "<th>Percent of<br>Total Bytes</th>\n",
  }
  print "</tr>\n";
  $direction = substr($interface_summary{"sort"},0,1);
  substr($interface_summary{"sort"},0,1) = "";
  if ($interface_summary{"sort"} eq "interface") {
    $sort = 'sort keys %total_interface_hits';
  } else {
    $sort =
      'sort { $total_interface_' . $interface_summary{"sort"} .
      '{$a} <=> $total_interface_' . $interface_summary{"sort"} .
      '{$b} } keys %total_interface_hits';
  }
  if ($direction eq "-") { $sort = 'reverse ' . $sort; }
  foreach $index (eval($sort)) {
    if ($highlight) { $class = $class eq "odd" ? "even" : "odd"; }
    print
      "<tr>\n",
      "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">$index</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">$total_interface_hits{$index}</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"left\">", fduration($total_interface_duration{$index}), "</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_interface_sent{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_interface_received{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_interface_total{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    printf
      " align=\"right\">%.2f</td>\n",
      100 * $total_interface_total{$index} / $total_total;
    if ($graphs) {
      print "<td";
      if ($highlight) { print " class=\"$class\""; }
      print ">";
      printf
        "<img alt=\"\" src=\"pixel.gif\" width=\"%d\" height=\"12\"></td>",
        width($total_interface_total{$index},$total_total);
    }
    print "</tr>\n";
    last if ($interface_summary{"limit"} && ++$num >= $interface_summary{"limit"});
  }
  print
    "</table>\n",
    "</p>\n";
}

sub host_summary {
  my ($host,$num,$direction,$sort,$class,$index);
  $host = shift;
  $num = 0;
  print
    "<p>\n",
    "<table $table>\n";
  if ($graphs) {
    print "<tr><th align=\"left\" colspan=\"8\">Host Summary [$host]</th></tr>\n";
  } else {
    print "<tr><th align=\"left\" colspan=\"7\">Host Summary [$host]</th></tr>\n";
  }
  print
    "<tr>\n",
    "<th>Host</th>\n",
    "<th>Hits</th>\n",
    "<th>Duration</th>\n",
    "<th>Bytes<br>Sent</th>\n",
    "<th>Bytes<br>Received</th>\n",
    "<th>Bytes<br>Total</th>\n";
  if ($graphs) {
    print "<th colspan=\"2\">Percent of Total Bytes</th>\n",
  } else {
    print "<th>Percent of<br>Total Bytes</th>\n",
  }
  print "</tr>\n";
  $direction = substr($host_summary{$host}{"sort"},0,1);
  substr($host_summary{$host}{"sort"},0,1) = "";
  if ($host_summary{$host}{"sort"} eq "host") {
    $sort = 'sort keys %{$total_host_hits{$host}}';
  } else {
    $sort =
      'sort { $total_host_' . $host_summary{$host}{"sort"} .
      '{$host}{$a} <=> $total_host_' . $host_summary{$host}{"sort"} .
      '{$host}{$b} } keys %{$total_host_hits{$host}}';
  }
  if ($direction eq "-") { $sort = 'reverse ' . $sort; }
  foreach $index (eval($sort)) {
    if ($highlight) { $class = $class eq "odd" ? "even" : "odd"; }
    print
      "<tr>\n",
      "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\"><nobr>", name($index), "</nobr></td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">$total_host_hits{$host}{$index}</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"left\">", fduration($total_host_duration{$host}{$index}), "</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_host_sent{$host}{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_host_received{$host}{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_host_total{$host}{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    printf
      " align=\"right\">%.2f</td>\n",
      100 * $total_host_total{$host}{$index} / $total_total;
    if ($graphs) {
      print "<td";
      if ($highlight) { print " class=\"$class\""; }
      print ">";
      printf
        "<img alt=\"\" src=\"pixel.gif\" width=\"%d\" height=\"12\"></td>",
        width($total_host_total{$host}{$index},$total_total);
    }
    print "</tr>\n";
    last if ($host_summary{$host}{"limit"} && ++$num >= $host_summary{$host}{"limit"});
  }
  print
    "</table>\n",
    "</p>\n";
}

sub user_summary {
  my ($num,$direction,$sort,$class,$index);
  $num = 0;
  if ($links) { makelinks("user_summary"); }
  print
    "<p>\n",
    "<table $table>\n";
  if ($graphs) {
    print "<tr><th align=\"left\" colspan=\"8\">User Summary</th></tr>\n";
  } else {
    print "<tr><th align=\"left\" colspan=\"7\">User Summary</th></tr>\n";
  }
  print
    "<tr>\n",
    "<th>User</th>\n",
    "<th>Hits</th>\n",
    "<th>Duration</th>\n",
    "<th>Bytes<br>Sent</th>\n",
    "<th>Bytes<br>Received</th>\n",
    "<th>Bytes<br>Total</th>\n";
  if ($graphs) {
    print "<th colspan=\"2\">Percent of Total Bytes</th>\n",
  } else {
    print "<th>Percent of<br>Total Bytes</th>\n",
  }
  print "</tr>\n";
  $direction = substr($user_summary{"sort"},0,1);
  substr($user_summary{"sort"},0,1) = "";
  if ($user_summary{"sort"} eq "user") {
    $sort = 'sort keys %total_user_hits';
  } else {
    $sort =
      'sort { $total_user_' . $user_summary{"sort"} . '{$a} <=> $total_user_' .
      $user_summary{"sort"} . '{$b} } keys %total_user_hits';
  }
  if ($direction eq "-") { $sort = 'reverse ' . $sort; }
  foreach $index (eval($sort)) {
    if ($highlight) { $class = $class eq "odd" ? "even" : "odd"; }
    print
      "<tr>\n",
      "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"left\"><nobr>$index</nobr></td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">$total_user_hits{$index}</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"left\">", fduration($total_user_duration{$index}), "</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_user_sent{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_user_received{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_user_total{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    printf
      " align=\"right\">%.2f</td>\n",
      100 * $total_user_total{$index} / $total_total;
    if ($graphs) {
      print "<td";
      if ($highlight) { print " class=\"$class\""; }
      print ">";
      printf
        "<img alt=\"\" src=\"pixel.gif\" width=\"%d\" height=\"12\"></td>",
        width($total_user_total{$index},$total_total);
    }
    print "</tr>\n";
    last if ($user_summary{"limit"} && ++$num >= $user_summary{"limit"});
  }
  print
    "</table>\n",
    "</p>\n";
}

sub hour_summary {
  my ($num,$direction,$sort,$class,$index);
  $num = 0;
  if ($links) { makelinks("hour_summary"); }
  print
    "<p>\n",
    "<table $table>\n";
  if ($graphs) {
    print "<tr><th align=\"left\" colspan=\"8\">Time of Day Summary</th></tr>\n";
  } else {
    print "<tr><th align=\"left\" colspan=\"7\">Time of Day Summary</th></tr>\n";
  }
  print
    "<tr>\n",
    "<th>Hour</th>\n",
    "<th>Hits</th>\n",
    "<th>Duration</th>\n",
    "<th>Bytes<br>Sent</th>\n",
    "<th>Bytes<br>Received</th>\n",
    "<th>Bytes<br>Total</th>\n";
  if ($graphs) {
    print "<th colspan=\"2\">Percent of Total Bytes</th>\n",
  } else {
    print "<th>Percent of<br>Total Bytes</th>\n",
  }
  print "</tr>\n";
  $direction = substr($hour_summary{"sort"},0,1);
  substr($hour_summary{"sort"},0,1) = "";
  if ($hour_summary{"sort"} eq "hour") {
    $sort = 'sort keys %total_hour_hits';
  } else {
    $sort =
      'sort { $total_hour_' . $hour_summary{"sort"} . '{$a} <=> $total_hour_' .
      $hour_summary{"sort"} . '{$b} } keys %total_hour_hits';
  }
  if ($direction eq "-") { $sort = 'reverse ' . $sort; }
  foreach $index (eval($sort)) {
    if ($highlight) { $class = $class eq "odd" ? "even" : "odd"; }
    print
      "<tr>\n",
      "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\"><nobr>$index:00 - $index:59</nobr></td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">$total_hour_hits{$index}</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"left\">", fduration($total_hour_duration{$index}), "</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_hour_sent{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_hour_received{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_hour_total{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    printf
      " align=\"right\">%.2f</td>\n",
      100 * $total_hour_total{$index} / $total_total;
    if ($graphs) {
      print "<td";
      if ($highlight) { print " class=\"$class\""; }
      print ">";
      printf
        "<img alt=\"\" src=\"pixel.gif\" width=\"%d\" height=\"12\"></td>",
        width($total_hour_total{$index},$total_total);
    }
    print "</tr>\n";
    last if ($hour_summary{"limit"} && ++$num >= $hour_summary{"limit"});
  }
  print
    "</table>\n",
    "</p>\n";
}

sub domain_summary {
  my ($num,$direction,$sort,$class,$index);
  $num = 0;
  if ($links) { makelinks("domain_summary"); }
  print
    "<p>\n",
    "<table $table>\n";
  if ($graphs) {
    print "<tr><th align=\"left\" colspan=\"8\">Top Level Domain Summary</th></tr>\n";
  } else {
    print "<tr><th align=\"left\" colspan=\"7\">Top Level Domain Summary</th></tr>\n";
  }
  print
    "<tr>\n",
    "<th>Domain</th>\n",
    "<th>Hits</th>\n",
    "<th>Duration</th>\n",
    "<th>Bytes<br>Sent</th>\n",
    "<th>Bytes<br>Received</th>\n",
    "<th>Bytes<br>Total</th>\n";
  if ($graphs) {
    print "<th colspan=\"2\">Percent of Total Bytes</th>\n",
  } else {
    print "<th>Percent of<br>Total Bytes</th>\n",
  }
  print "</tr>\n";
  $direction = substr($domain_summary{"sort"},0,1);
  substr($domain_summary{"sort"},0,1) = "";
  if ($hour_summary{"sort"} eq "domain") {
    $sort = 'sort keys %total_domain_hits';
  } else {
    $sort =
      'sort { $total_domain_' . $domain_summary{"sort"} .
      '{$a} <=> $total_domain_' . $domain_summary{"sort"} .
      '{$b} } keys %total_domain_hits';
  }
  if ($direction eq "-") { $sort = 'reverse ' . $sort; }
  foreach $index (eval($sort)) {
    if ($highlight) { $class = $class eq "odd" ? "even" : "odd"; }
    print
      "<tr>\n",
      "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">$index</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">$total_domain_hits{$index}</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"left\">", fduration($total_domain_duration{$index}), "</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_domain_sent{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_domain_received{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">", bytes($total_domain_total{$index}) ,"</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    printf
      " align=\"right\">%.2f</td>\n",
      100 * $total_domain_total{$index} / $total_total;
    if ($graphs) {
      print "<td";
      if ($highlight) { print " class=\"$class\""; }
      print ">";
      printf
        "<img alt=\"\" src=\"pixel.gif\" width=\"%d\" height=\"12\"></td>",
        width($total_domain_total{$index},$total_total);
    }
    print "</tr>\n";
    last if ($domain_summary{"limit"} && ++$num >= $domain_summary{"limit"});
  }
  print
    "</table>\n",
    "</p>\n";
}

sub mobile_summary {
  my ($num,$direction,$sort,$class,$index);
  $num = 0;
  if ($links) { makelinks("mobile_summary"); }
  print
    "<p>\n",
    "<table $table>\n";
  if ($graphs) {
    print "<tr><th align=\"left\" colspan=\"3\">Raptor Mobile Summary</th></tr>\n",
  } else {
    print "<tr><th align=\"left\" colspan=\"2\">Raptor Mobile Summary</th></tr>\n",
  }
  print
    "<tr>\n",
    "<th>Net Entity</th>\n";
  if ($graphs) {
    print "<th colspan=\"2\">Connections</th>\n";
  } else {
    print "<th>Connections</th>\n";
  }
  print
    "</tr>\n";
  $direction = substr($mobile_summary{"sort"},0,1);
  substr($mobile_summary{"sort"},0,1) = "";
  if ($mobile_summary{"sort"} eq "host") {
    $sort = 'sort keys %total_mobile';
  } else {
    $sort = 'sort { $total_mobile{$a} <=> $total_mobile{$b} } keys %total_mobile';
  }
  if ($direction eq "-") { $sort = 'reverse ' . $sort; }
  foreach $index (eval($sort)) {
    if ($highlight) { $class = $class eq "odd" ? "even" : "odd"; }
    print
      "<tr>\n",
      "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\"><nobr>", name($index), "</nobr></td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">$total_mobile{$index}</td>\n";
    if ($graphs) {
      print "<td";
      if ($highlight) { print " class=\"$class\""; }
      print ">";
      printf
        "<img alt=\"\" src=\"pixel.gif\" width=\"%d\" height=\"12\"></td>",
        width($total_mobile{$index},$total_mobile);
    }
    print
      "</tr>\n";
    last if ($mobile_summary{"limit"} && ++$num >= $mobile_summary{"limit"});
  }
  print
    "</table>\n",
    "</p>\n";
}

sub message_summary {
  my ($num,$direction,$sort,$class,$index);
  if ($links) { makelinks("message_summary"); }
  print
    "<p>\n",
    "<table $table>\n",
    "<tr><th align=\"left\" colspan=\"4\">Message Summary</th></tr>\n",
    "<tr>\n",
    "<th>Message</th>\n",
    "<th>Severity</th>\n",
    "<th>Hits</th>\n",
    "<th>Description</th>\n",
    "</tr>\n";
  $direction = substr($message_summary{"sort"},0,1);
  substr($message_summary{"sort"},0,1) = "";
  if ($message_summary{"sort"} eq "message") {
    $sort = 'sort keys %total_message';
  } else {
    $sort = 'sort { $total_message{$a} <=> $total_message{$b} } keys %total_message';
  }
  if ($direction eq "-") { $sort = 'reverse ' . $sort; }
  foreach $index (eval($sort)) {
    if ($highlight) { $class = $class eq "odd" ? "even" : "odd"; }
    print
      "<tr>\n",
      "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\"><nobr>$index</nobr></td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print
      " align=\"left\">",
      $message_severity[int($index / 100) - 1],
      "</td>\n",
      "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"right\">$total_message{$index}</td>\n";
    print "<td";
    if ($highlight) { print " class=\"$class\""; }
    print " align=\"left\"><nobr>$message_description{$index}</nobr></td>\n";
    print "</tr>\n";
  }
  print
    "</table>\n",
    "</p>\n";
}

sub history_summary {
  my ($value,$num,$base,$index,$max,@history_stats);
  $value = shift;
  $num = 0;
  $max = 0;
  unless (open HISTORY, $history_file) {
    print STDERR
      "WARNING: Unable to open history file. This is not a fatal error,\n",
      "processing will continue. If this is not the first time you are\n",
      "running Reptor with the history_file option specified, then you will\n",
      "probably need to include the full path along with the name of the\n",
      "history file.\n";
    return;
  }
  while (<HISTORY>) {
    my ($history_day,$history_hits,$history_total,$history_duration);
    ($history_day,$history_hits,$history_total,$history_duration) = split(/,/);
    $history_stats[$num]{"day"} = $history_day;
    $history_stats[$num]{"hits"} = $history_hits;
    $history_stats[$num]{"total"} = $history_total;
    $history_stats[$num]{"duration"} = $history_duration;
    $num++;
  }
  close HISTORY;
  $base = $num > $history_summary{$value} ? $num - $history_summary{$value} : 0;
  for ($index = 0; $index < $history_summary{$value}; $index++) {
    if ($history_stats[$base + $index]{$value} > $max) {
      $max = $history_stats[$base + $index]{$value};
    }
  }
  print
    "<p>\n",
    "<table $table>\n",
    "<tr><th align=\"left\" colspan=\"", $history_summary{$value} + 1,
    "\">History Summary [$value]</th></tr>\n",
    "<tr>\n",
    "<td valign=\"top\">",
    ($value eq "total") ? bytes($max) : $max,
    "</td>\n";
  for ($index = 0; $index < $history_summary{$value}; $index++) {
    print "<td align=\"center\" valign=\"bottom\">";
    if ($graphs) {
       print
        "<img alt=\"\" src=\"pixel.gif\" width=\"10\" height=\"",
        height($history_stats[$base + $index]{$value},$max),
        "\">";
    }
    print "</td>\n";
  }
  print
    "</tr>\n",
    "<tr>\n",
    "<td></td>\n";
  for ($index = $history_summary{$value} - 1; $index >= 0; $index--) {
    printf "<td>%02d</td>\n", $index;
  }
  print
    "</tr>\n",
    "<tr>\n",
    "<td align=\"center\" colspan=\"", $history_summary{$value} + 1,
    "\">Days Ago</td>\n",
    "</tr>\n",
    "</table>\n",
    "</p>\n";
}

sub bytes {
  my $bytes = shift;
  if ($nice_bytes) {
    if ($bytes >= 1024000) {
      return sprintf "<nobr>%.2f Mb</nobr>", $bytes / 1048576;
    }
    if ($bytes >= 1000) {
      return sprintf "<nobr>%.1f Kb</nobr>", $bytes / 1024;
    }
  }
  return $bytes;
}

sub fduration {
  my ($duration,$hour,$minute,$second);
  $duration = shift;
  $hour = int($duration / 3600);
  $duration -= ($hour * 3600);
  $minute = int($duration / 60);
  $duration -= ($minute * 60);
  $second = $duration;
  return sprintf "%03d:%02d:%02d", $hour, $minute, $second;
}

sub height {
  my ($part,$whole,$height);
  $part = shift;
  $whole = shift;
  $height = $graphs * 500 * $part / $whole;
  return $height < 1 ? 1 : int($height + 0.5);
}

sub width {
  my ($part,$whole,$width);
  $part = shift;
  $whole = shift;
  $width = $graphs * 1000 * $part / $whole;
  return $width < 1 ? 1 : int($width + 0.5);
}

sub name {
  my $ip = shift;
  if ($dns eq "print" && $ip =~ /^[0-9]+\.[0-9]+\.[0-9]+\.[0-9]+$/) {
    return lc(gethostbyaddr(pack("c4",split(/\./,$ip)),2)) || "[$ip]";
  } else {
    return $ip;
  }
}

sub makelinks {
  my $name = shift;
  print "<p><a name=\"$name\"></a>\n";
  if ($alerts) {
    print "[<a href=\"#alerts\">Alerts</a>] \n";
  }
  if (%alert_summary) {
    print "[<a href=\"#alert_summary\">Alert</a>] \n";
  }
  if ($totals) {
    print "[<a href=\"#totals\">Totals</a>] \n";
  }
  if (%protocol_summary) {
    print "[<a href=\"#protocol_summary\">Protocol</a>] \n";
  }
  if (%hour_summary) {
    print "[<a href=\"#hour_summary\">Time of Day</a>] \n";
  }
  if (%interface_summary) {
    print "[<a href=\"#interface_summary\">Interface</a>] \n";
  }
  if (%host_summary) {
    print "[<a href=\"#host_summary\">Host</a>]\n";
  }
  if (%user_summary) {
    print "[<a href=\"#user_summary\">User</a>]\n";
  }
  if (%domain_summary) {
    print "[<a href=\"#domain_summary\">Domain</a>]\n";
  }
  if (%mobile_summary) {
    print "[<a href=\"#mobile_summary\">Mobile</a>]\n";
  }
  if (%message_summary) {
    print "[<a href=\"#message_summary\">Message</a>]\n";
  }
  if (%history_summary) {
    print "[<a href=\"#history_summary\">History</a>]\n";
  }
  print "</p>\n";
}

sub config {

  my ($config_file,$option,$arg1,$arg2,$arg3);

  $config_file = shift;

  open CONFIG, $config_file or
    die "ERROR: Unable to open configuration file $config_file.\n";

  while (<CONFIG>) {
    next if /^#/ || /^$/;

    ($option,$arg1,$arg2,$arg3) = /^\s*([a-zA-Z_]+)(?:\s+([^\s\"]+|(?:\"[^\"]+\")))?(?:\s+([^\s\"]+|(?:\"[^\"]+\")))?(?:\s+([^\s\"]+|(?:\"[^\"]+\")))?\s+$/;

    if (!$option) {
      die "ERROR: Bad format in configuration file:\n  $_";
    }

    $option = lc($option);
    $arg1 =~ tr/\"//d;
    $arg2 =~ tr/\"//d;
    $arg3 =~ tr/\"//d;

    SWITCH: {

      if ($option =~ /^(directory|compress|history_file|mail_server|mail_from|dns|title|style_sheet|table|graphs)$/) {
        $$option = $arg1;
        last SWITCH;
      }

      if ($option eq "remotelog") {
        $remotelog = $arg1;
        if ($arg2) { $remotelog_path = $arg2; }
        last SWITCH;
      }

      if ($option eq "save_logfile") {
        $save_logfile = $arg1 || ".";
        last SWITCH;
      }

      if ($option eq "uncompress") {
        $uncompress = $arg1;
        $extension = $arg2;
        last SWITCH;
      }

      if ($option eq "interface") {
        $interface{$arg1} = $arg2 || $arg1;
        last SWITCH;
      }

      if ($option eq "filter") {
        unless ($arg1 eq "any" && $arg2 eq "any") {
          push @{$filter{"source"}}, $arg1;
          push @{$filter{"destination"}}, $arg2;
        }
        last SWITCH;
      }

      if ($option eq "alias") {
        $alias{$arg1} = $arg2;
        last SWITCH;
      }

      if ($option eq "time") {
        push @{$time{$arg1}{"begin"}}, $arg2;
        push @{$time{$arg1}{"end"}}, $arg3;
        last SWITCH;
      }

      if ($option eq "duration") {
        $arg3 = lc($arg3);
        if ($arg3 eq "h") { $arg3 = 3600; }
        elsif ($arg3 eq "m") { $arg3 = 60; }
        else { $arg3 = 1; }
        $duration{$arg1} = $arg2 * $arg3;
        last SWITCH;
      }

      if ($option eq "volume") {
        $arg3 = lc($arg3);
        if ($arg3 eq "m") { $arg3 = 1048576; }
        elsif ($arg3 eq "k") { $arg3 = 1024; }
        else { $arg3 = 1; }
        $volume{$arg1} = $arg2 * $arg3;
        last SWITCH;
      }

      if ($option eq "user") {
        $user{$arg1}{$arg2} = 1;
        last SWITCH;
      }

      if ($option eq "word") {
        $word .= $arg1 . "|";
        last SWITCH;
      }

      if ($option =~ /^(alert|protocol|hour|interface|user|domain|message|mobile)_summary$/) {
        ${$option}{"sort"} = $arg1 =~ /^[a-zA-Z]/ ? "+" . $arg1 : $arg1;
        ${$option}{"limit"} = $arg2;
        last SWITCH;
      }

      if ($option eq "history_summary") {
        $history_summary{$arg1} = $arg2;
        last SWITCH;
      }

      if ($option eq "host_summary") {
        $host_summary{$arg1}{"sort"} =
          $arg2 =~ /^[a-zA-Z]/ ? "+" . $arg2 : $arg2;
        $host_summary{$arg1}{"limit"} = $arg3;
        last SWITCH;
      }

      if ($option eq "output") {
        $output_how = $arg1;
        $output_to = $arg2;
        last SWITCH;
      }

      $$option = 1;

    }
  }

  close CONFIG;

  # set defaults
  $output_how ||= "stdout";
  $title ||= "Firewall activity report";
  $table ||= "border=\"0\" cellpadding=\"2\" cellspacing=\"2\"";
  $mail_server ||= "127.0.0.1";
  $mail_from ||= "Reptor";
  if (%message_summary) { $message_summary{"limit"} ||= 1000; }

  # add a slash to remotelogpath if one is not already there
  if ($remotelog_path) { $remotelog_path =~ s/([^\/\\])$/$1\//; }

  # check validity of specified options
  if ($alerts && !(%time || %duration || %volume || %user || $ratings || $word)) {
    die "ERROR: Option \"alerts\" requires at least one of:\n  \"time\", \"duration\", \"volume\", \"user\", \"ratings\", or \"word\".\n";
  }

  if ($verify && $show_reason && !$alerts) {
    die "WARNING: Option \"show_reason\" requires \"alerts\".\n";
  }

  if ($verify && $show_user && !$alerts) {
    die "WARNING: Option \"show_user\" requires \"alerts\".\n";
  }

  if ($verify && $show_auth && !$alerts) {
    die "WARNING: Option \"show_auth\" requires \"alerts\".\n";
  }

  if ($verify && $show_rule && !$alerts) {
    die "WARNING: Option \"show_rule\" requires \"alerts\".\n";
  }

  if ($verify && $show_id && !$alerts) {
    die "WARNING: Option \"show_id\" requires \"alerts\".\n";
  }

  if ($verify && $show_op && !$alerts) {
    die "WARNING: Option \"show_op\" requires \"alerts\".\n";
  }

  if ($verify && $show_result && !$alerts) {
    die "WARNING: Option \"show_result\" requires \"alerts\".\n";
  }

  if ($verify && $show_arg && !$alerts) {
    die "WARNING: Option \"show_arg\" requires \"alerts\".\n";
  }

  if ($verify && $all && !$alerts) {
    die "WARNING: Option \"all\" requires \"alerts\".\n";
  }

  if ($verify && %time && !$alerts) {
    die "WARNING: Option \"time\" requires \"alerts\".\n";
  }

  if ($verify && %duration && !$alerts) {
    die "WARNING: Option \"duration\" requires \"alerts\".\n";
  }

  if ($verify && %volume && !$alerts) {
    die "WARNING: Option \"volume\" requires \"alerts\".\n";
  }

  if ($verify && %user && !$alerts) {
    die "WARNING: Option \"user\" requires \"alerts\".\n";
  }

  if ($verify && $word && !$alerts) {
    die "WARNING: Option \"word\" requires \"alerts\".\n";
  }

  if ($word && !($search_source || $search_destination || $search_arg)) {
    die "ERROR: Option \"word\" requires at least one of:\n \"search_source\", \"search_destination\", or \"search_arg\".\n";
  }

  if ($verify && $search_source && !$word) {
    die "WARNING: Option \"search_source\" requires \"word\".\n";
  }

  if ($verify && $search_destination && !$word) {
    die "WARNING: Option \"search_destination\" requires \"word\".\n";
  }

  if ($verify && $search_arg && !$word) {
    die "WARNING: Option \"search_arg\" requires \"word\".\n";
  }

  if (%alert_summary && !$alerts) {
    die "ERROR: Option \"alert_summary\" requires \"alerts\".\n";
  }

  if ($output_how !~ /^(file|ftp|mail|stdout)$/) {
    die "ERROR: Option \"output\" must be one of \"stdout\", \"file\", \"ftp\", or \"mail\".\n";
  }

  if (!$totals && !$alerts && !%alert_summary && !%protocol_summary && !%host_summary && !%user_summary && !%hour_summary && !%domain_summary && !%mobile_summary && !%message_summary && !%history_summary) {
    die "ERROR: Selected options will produce no output.\n";
  }

  if ($verify && $remotelog && (! -x $remotelog_path . ($^O eq "MSWin32" ? "remotelogfile.exe" : "remotelogfile"))) { 
    die "ERROR: remotelogfile program not found.\n";
  }

  if ($verify && $save_logfile && !$remotelog) {
    die "WARNING: It doesn't make much sense to use save_logfile without using remotelog.\n";
  }

  if ($save_logfile && ! -d $save_logfile) {
    die "Directory specified by save_logfile does not exist.\n";
  }

  if ($verify && $compress && !$save_logfile) {
    die "WARNING: It doesn't make much sense to use compress without using save_logfile.\n";
  }

  if ($uncompress && ! -x (split(/ /,$uncompress))[0]) {
    die "ERROR: Unable to run uncompression program.\n";
  }

  if ($verify && $highlight && !$style_sheet) {
    die "WARNING: Must specify style_sheet with highlight.\n";
  }

  if ($verify && %history_summary && !$history_file) {
    die "WARNING: Must specify history_file with history_summary.\n";
  }

  if ($verify && %history_summary && !$graphs) {
    die "WARNING: Must specify graphs with history_summary.\n";
  }

  if ($verify) { print "Configuration file parsed with no errors.\n"; }

  # make a proper regular expression
  if ($word) {
    chop($word);
    $word = "($word)";
  }

  # set number of columns in alert section
  if ($alerts) {
    $columns = 8 + (2 * $show_interface) + $show_reason +
      $show_user + $show_auth + $show_rule + $show_id +
      $show_op + $show_result + $show_arg;
  }

}

sub version {
  print << "EOF";
Reptor version $reptor_version, Copyright (C) 1999,2000 Alex Howansky
Visit the Reptor home page at http://www.wankwood.com/reptor/
EOF
  exit;
}

sub usage {
  print << "EOF";
Reptor version $reptor_version, Copyright (C) 1999,2000 Alex Howansky

Reptor is distributed under the GNU GPL and provided without
any warranty. For additional information, refer to the file
"COPYING" that has been supplied with this distribution.

Usage: reptor.pl [options]

Options:

  --config <file>
    Name of configuration file.
    Default: reptor.cfg

  --date <date>
    Date of logfile to process.
    Default: $yesterday
    Note: v4.0 format is YYMMDD, v5.0 and v6.0 format is YYYYMMDD.

  --dir <directory>
    Directory where logfile is located.
    Default: $directory
    Note: Overrides "directory" configuration file option.

  --help

  --ignore
    The log to process has been cut from a larger file or merged from multiple
    files, so don't look for the datestamp on the first line, or ignore it if
    it's there.

  --log <file>
    Name of logfile to process.
    Default: $directory/logfile.$yesterday
    Note: Overrides --date and --directory command line options,
          and "directory" configuration file option.
          Use "--log -" to indicate STDIN input.

  --verify
    Verify validity of configuration file, readability of logfile, and exit
    without processing.

  --version
    Print version number and exit.

EOF
  exit;
}
